# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module containing enumeration classes for the jax.cs.storage package."""

from enum import Enum


class UrlTypes(Enum):
    """An enumeration of url types specific to this package."""

    HTTPS = 'https'
    R2 = 'r2'
    GCS = 'gs'
    FILE_IO = 'local_file'
    LOCAL_MEMORY = 'local_memory'
